@@ -6,8 +6,11 @@ from .api import (  # noqa
     ClientError,
     GCSPath,
     PureGCSPath,
+    clear_fs_cache,
+    get_fs_cache,
     get_fs_client,
     use_fs,
+    use_fs_cache,
 )
 from .client import BucketEntry, BucketStat  # noqa
 from .file import BucketClientFS, BucketEntryFS, ClientBlobFS, ClientBucketFS  # noqa
